//
//  NLURLBlockProvider.m
//  NuDataProvider
//
//  Created by Remy Demarest on 24/11/2010.
//  Copyright 2010 NuLayer Inc. All rights reserved.
//

#import "NLURLParser.h"

typedef enum _NLPathType
{
    NLPathTypePlain  = 0,
    NLPathTypePrefix = 1 << 0,
    NLPathTypeSuffix = 1 << 1,
    NLPathTypeInfix  = NLPathTypePrefix | NLPathTypeSuffix,
} NLPathType;

@interface NLURLParser ()
- (NLPathType)NL_pathTypeForPath:(NSString *)aPath getRawPath:(NSString **)resultPath;
@end


@implementation NLURLParser
@synthesize copyObjects = _copyObjects;
- (id)init
{
    return [self initWithCopyObjects:YES];
}

- (id)initWithCopyObjects:(BOOL)copy;
{
    self = [super init];
    if(self != nil)
    {
        _copyObjects = copy;
        
        _schemeToObjects       = [[NSMutableDictionary alloc] init];
        _pathsToObjects        = [[NSMutableDictionary alloc] init];
        _pathPrefixesToObjects = [[NSMutableDictionary alloc] init];
        _pathSuffixesToObjects = [[NSMutableDictionary alloc] init];
        _pathInfixesToObjects  = [[NSMutableDictionary alloc] init];
    }
    return self;
}

// URL Parser blocks for NSFetchRequest creation
- (id)objectForURL:(NSURL *)aURL;
{
    id ret = nil;
    
    ret = [_schemeToObjects objectForKey:[[aURL scheme] lowercaseString]];
    
    if(ret == nil) ret = [self objectForPath:([aURL path] ? : [aURL resourceSpecifier])];
    
    if(ret == nil) ret = _defaultObject;
    
    return ret;
}

- (id)objectForPath:(NSString *)aPath;
{
    id(^ret)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError) = nil;
    
    aPath = [aPath lowercaseString];
    
    // Search full paths
    ret = [_pathsToObjects objectForKey:aPath];
    
#define FIND_OBJECT(test, collection) do {       \
    __typeof__(collection) coll = (collection);  \
    if(ret == nil)                               \
        for(NSString *value in coll)             \
            if(test)                             \
            {                                    \
                ret = [coll objectForKey:value]; \
                break;                           \
            }                                    \
} while(0)
    
    FIND_OBJECT([aPath hasPrefix:value], _pathPrefixesToObjects);
    FIND_OBJECT([aPath hasSuffix:value], _pathSuffixesToObjects);
    FIND_OBJECT([aPath rangeOfString:value].location != NSNotFound, _pathInfixesToObjects);
    
    return ret;
}

- (NLPathType)NL_pathTypeForPath:(NSString *)aPath getRawPath:(NSString **)resultPath;
{
    NLPathType ret = NLPathTypePlain;
    
    NSRange pathRange = NSMakeRange(0, [aPath length]);
    
    if([aPath hasPrefix:@"*"])
    {
        // The string represent is a path suffix
        ret |= NLPathTypeSuffix;
        
        // Remove the wildcard from the path
        pathRange.location++;
        pathRange.length--;
    }
    
    if([aPath hasSuffix:@"*"] && ! [aPath hasPrefix:@"\\*"])
    {
        // The string represent is a path prefix
        ret |= NLPathTypePrefix;
        
        // Remove the wildcard from the path
        pathRange.length--;
    }
    
    if(resultPath != NULL)
    {
        aPath = [aPath substringWithRange:pathRange];
        
        // Replace escaped \\* with *
        aPath = [aPath stringByReplacingOccurrencesOfString:@"\\*" withString:@"*"];
        
        // Replace escaped \\\\ with \\
        aPath = [aPath stringByReplacingOccurrencesOfString:@"\\\\" withString:@"\\"];
        
        *resultPath = [aPath lowercaseString];
    }
    
    return ret;
}

// Schemes, no wildcards
- (void)setObject:(id)obj forURLScheme:(NSString *)aScheme;
{
    aScheme = [aScheme lowercaseString];
    if([self copyObjects]) obj = [obj copy];
    
    if(obj != nil) [_schemeToObjects setObject:obj forKey:aScheme];
    else           [_schemeToObjects removeObjectForKey:aScheme];
}
- (void)removeObjectForURLScheme:(NSString *)aScheme;
{
    [self setObject:nil forURLScheme:aScheme];
}

// All paths are lowercase
// The path accept wildcards as follow: "prefix/*" "*/infix/*" "*/suffix" 
- (void)setObject:(id)obj forURLPath:(NSString *)aPath;
{
    if([self copyObjects]) obj = [obj copy];
    
    NSMutableDictionary *targeted = nil;
    
    switch([self NL_pathTypeForPath:aPath getRawPath:&aPath])
    {
        case NLPathTypePlain  : targeted = _pathsToObjects       ; break;
        case NLPathTypePrefix : targeted = _pathPrefixesToObjects; break;
        case NLPathTypeSuffix : targeted = _pathSuffixesToObjects; break;
        case NLPathTypeInfix  : targeted = _pathInfixesToObjects ; break;
    }
    
    if(obj != nil) [targeted setObject:obj forKey:aPath];
    else           [targeted removeObjectForKey:aPath];
}

- (void)removeObjectForURLPath:(NSString *)aPath;
{
    [self setObject:nil forURLPath:aPath];
}

- (id)defaultObject; { return _defaultObject; }
- (void)setDefaultObject:(id)value;
{
    if(_defaultObject != value)
    {
        _defaultObject = ([self copyObjects] ? [value copy] : value);
    }
}

@end
