//
//  NLImageProvider.m
//  NLDataProvider
//
//  Created by Remy Demarest on 15/10/2010.
//  Copyright 2010 Nulayer Inc. All rights reserved.
//

#import "NLImageProvider.h"
#import <NLFoundation/NLFoundation.h>

static NSString *const NLImageLastAccessedDate = @"NLImageLastAccessedDate";
static NSString *const NLImageLastModifiedDate = @"NLImageLastModifiedDate";

@interface NLImageProviderHandler : NSObject
{
@private
    id            _observer;
    NSMutableSet *_loaders;
}
+ (id)handler;
@property(copy, nonatomic, readonly) NSMutableSet *loaders;
@property(nonatomic)                 id            observer;
@property(nonatomic)                 BOOL          continueLoading;
@end

@interface NLImageProvider () <NLDataProviderRequestDelegate>
{
    NSString *_cachePath;
    NSMutableDictionary *_imagesDictionary;
}

- (NSString *)directoryPathForURL:(NSURL *)aURL;
- (NSString *)filePathForURL:(NSURL *)aURL;

- (void)updateImageInPlist:(NSString *)imagePath withLastModifiedDate:(NSString *)aDate;
- (void)writePlistToFile;

@end

static const NSInteger MAX_IMAGE_CACHE_SIZE = 26214400; // Limit the image cache size to 25 megs

@implementation NLImageProvider
@synthesize dataProvider = _dataProvider;

static NLImageProvider *_sharedImageProvider = nil;

+ (void)initialize
{
    if(self == [NLImageProvider class])
    {
        _sharedImageProvider = [[self alloc] init];
    }
}

+ (NLImageProvider *)sharedImageProvider;
{
    return _sharedImageProvider;
}

- (NSMutableDictionary *)imagesDictionary
{
    if(_imagesDictionary == nil)
    {
        if(![[NSFileManager defaultManager] fileExistsAtPath:[self plistPath]])
            _imagesDictionary = [[NSMutableDictionary alloc] init];
        else
            _imagesDictionary = [[NSMutableDictionary alloc] initWithContentsOfFile:[self plistPath]];
    }
    
    return _imagesDictionary;
}

- (id)init
{
    self = [super init];
    if(self != nil)
    {
        _loaderToURL  = [[NSMutableDictionary alloc] init];
        _imageCache   = [[NSCache alloc] init];
        _imageLoaders = [[NSMutableDictionary alloc] init];
    }
    return self;
}

- (id)initWithCachePath:(NSString *)aPath;
{
    if(self = [super init])
    {
        _cachePath = [aPath copy];
        [[NSFileManager defaultManager] createDirectoryAtPath:_cachePath withIntermediateDirectories:YES attributes:nil error:NULL];
    }
    
    return self;
}

- (void)setCachePath:(NSString *)aPath
{
    if(_cachePath != nil)
        return;
    
    _cachePath = [aPath copy];
    [[NSFileManager defaultManager] createDirectoryAtPath:_cachePath withIntermediateDirectories:YES attributes:nil error:NULL];
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

#pragma mark -
#pragma mark Image loaders

- (void)NL_returnImage:(UIImage *)anImage forURL:(NSURL *)aURL toLoaders:(NSSet *)loaders
{
    for(id loader in loaders)
    {
        if([loader respondsToSelector:@selector(imageProvider:didFinishLoadingImage:forURL:)])
            [loader imageProvider:self didFinishLoadingImage:anImage forURL:aURL];
        else if([NSStringFromClass([loader class]) rangeOfString:@"Block"].length != 0)
            ((void(^)(id resources, NSDictionary *extraInfo, NSError *error))loader)(anImage, nil, nil);
        
        [_loaderToURL removeObjectForKey:[NSValue valueWithNonretainedObject:loader]];
    }
}

- (NLImageProviderHandler *)NL_createHandlerForURL:(NSURL *)aURL
{
    if(_dataProvider == nil) return nil;
    
    NLImageProviderHandler *ret = [NLImageProviderHandler handler];
    
    [_imageLoaders setObject:ret forKey:aURL];
    
    NSDictionary *optionsDictionary = nil;
    BOOL wasCached = NO;
        
    if(_cachePath != nil)
    {
        NSString *lastModifiedDate = [[[self imagesDictionary] objectForKey:[self filePathForURL:aURL]] objectForKey:NLImageLastModifiedDate];
        
        optionsDictionary = lastModifiedDate ? [NSDictionary dictionaryWithObject:[NSDictionary dictionaryWithObject:lastModifiedDate forKey:@"If-Modified-Since"] forKey:NLDataProviderHTTPHeadersOptions] : nil;
        
        wasCached = lastModifiedDate != nil;
    }
    
    id observer =
    [_dataProvider loadResourcesForURL:aURL options:optionsDictionary delegate:self resourceParser:nil completion:
     ^(id resources, NSDictionary *extraInfo, NSError *error)
     {
         if(![resources isKindOfClass:[UIImage class]])
             return;
         
         if(resources != nil)
         {
            if(_cachePath == nil)
                [_imageCache setObject:resources forKey:aURL cost:MAX([[extraInfo objectForKey:NLDataProviderNSURLResponseExtraInfo] expectedContentLength], 0)];
            else
            {                
                NSString *directoryPath    = [self directoryPathForURL:aURL];
                NSString *filePath         = [self filePathForURL:aURL];
                NSFileManager *fileManager = [NSFileManager defaultManager];
                
                // if we receive an image that we already have, overwrite it
                if([fileManager fileExistsAtPath:filePath])
                    [fileManager removeItemAtPath:filePath error:NULL];  
                else
                    [fileManager createDirectoryAtPath:[_cachePath stringByAppendingPathComponent:directoryPath] withIntermediateDirectories:YES attributes:nil error:NULL];

                [fileManager createFileAtPath:[_cachePath stringByAppendingPathComponent:filePath] contents:UIImagePNGRepresentation(resources) attributes:nil];
                
                [self updateImageInPlist:filePath withLastModifiedDate:[[extraInfo objectForKey:@"allHeaderFields"] objectForKey:@"Last-Modified"]];
            }
         }
         
         [_imageLoaders removeObjectForKey:aURL];
         
         if(!wasCached)
             [self NL_returnImage:resources forURL:aURL toLoaders:[ret loaders]];
         else if(wasCached && resources == nil)
         {
             [self updateImageInPlist:[self filePathForURL:aURL] withLastModifiedDate:[[extraInfo objectForKey:@"allHeaderFields"] objectForKey:@"Last-Modified"]];
         }
         
         [ret setObserver:nil];
     }];
    
    [ret setObserver:observer];
    
    return ret;
}

- (void)loadImageAtURL:(NSURL *)aURL forLoader:(id<NLImageProviderLoader>)loader;
{
    [self loadImageAtURL:aURL forLoader:loader finishLoading:NO];
}

- (void)loadImageAtURL:(NSURL *)aURL forLoader:(id<NLImageProviderLoader>)loader finishLoading:(BOOL)alwaysFinish;
{
    if(aURL == nil || [aURL isEqual:[NSNull null]])
    {
        [loader imageProvider:self didFinishLoadingImage:nil forURL:aURL];
        return;
    }
    
    UIImage *content = nil;
    
    // if we aren't using the disk cache, use the memory cached image
    if(_cachePath == nil)
    {
        content = [_imageCache objectForKey:aURL];
    
        if(content != nil) return [loader imageProvider:self didFinishLoadingImage:content forURL:aURL];
    }
    
    NLImageProviderHandler *handler = [_imageLoaders objectForKey:aURL];
    
    if(handler == nil)
    {
        handler = [self NL_createHandlerForURL:aURL];
        if(handler == nil) return [loader imageProvider:self didFinishLoadingImage:nil forURL:aURL];
    }
    
    if(alwaysFinish) [handler setContinueLoading:alwaysFinish];
    
    [[handler loaders] addObject:loader];
    [_loaderToURL setObject:aURL forKey:[NSValue valueWithNonretainedObject:loader]];
    
    
    // return the disk cached image immediately
    UIImage *image = [UIImage imageWithContentsOfFile:[_cachePath stringByAppendingPathComponent:[self filePathForURL:aURL]]];
    
    if(image != nil) return [loader imageProvider:self didFinishLoadingImage:image forURL:aURL];
}

- (void)stopLoadingImageForLoader:(id<NLImageProviderLoader>)loader;
{
    NSValue *loaderKey = [NSValue valueWithNonretainedObject:loader];
    
    NSURL *aURL = [_loaderToURL objectForKey:loaderKey];
    
    if(aURL != nil)
    {
        NLImageProviderHandler *handler = [_imageLoaders objectForKey:aURL];
        
        if(handler != nil)
        {
            [[handler loaders] removeObject:loader];
            
            if(![handler continueLoading] && [[handler loaders] count] == 0)
            {
                [_dataProvider removeCompletionBlock:[handler observer]];
                [_imageLoaders removeObjectForKey:aURL];
            }
        }
        
        [_loaderToURL removeObjectForKey:loaderKey];
    }
}

#pragma mark -
#pragma mark Utilities

- (unsigned long long)sizeOfDiskCache;
{    
    NSDirectoryEnumerator *e = [[NSFileManager defaultManager] enumeratorAtPath:_cachePath];
    
    unsigned long long cacheSize = 0;
    
    while([e nextObject])
        cacheSize += [[[e fileAttributes] valueForKey:NSFileSize] unsignedLongLongValue];
    
    return cacheSize;
}

- (NSString *)directoryPathForURL:(NSURL *)aURL
{
    return [[aURL path] stringByDeletingLastPathComponent];
}

- (NSString *)filePathForURL:(NSURL *)aURL
{
    return [aURL path];
}

- (void)purgeOldImagesFromCache;
{
    NSArray *allKeys = [[[self imagesDictionary] allKeys] copy];
    
    // Remove images from the cache that haven't been accessed in a week
    for(NSString *key in allKeys)
    {
        if(-[[[[self imagesDictionary] valueForKey:key] objectForKey:NLImageLastAccessedDate] timeIntervalSinceNow] > 3600.0 * 24.0 * 7.0)
        {
            [[NSFileManager defaultManager] removeItemAtPath:[_cachePath stringByAppendingPathComponent:key] error:NULL];
            [[self imagesDictionary] removeObjectForKey:key];
        }
    }
    
    // If the size of the disk cache exceeds the maximum image cache size, remove old images until no longer the case
    if([self sizeOfDiskCache] > MAX_IMAGE_CACHE_SIZE)
    {        
        NSArray *keysSortedByDates = [[[self imagesDictionary] keysSortedByValueUsingComparator:
        ^NSComparisonResult(id obj1, id obj2)
         {
             return [[obj1 objectForKey:NLImageLastAccessedDate] compare:[obj2 objectForKey:NLImageLastAccessedDate]];
         }] copy];
        
        unsigned long long currentDiskCacheSize = [self sizeOfDiskCache];
        
        for(NSString *key in keysSortedByDates)
        {
            currentDiskCacheSize -= [[[[NSFileManager defaultManager] attributesOfItemAtPath:[_cachePath stringByAppendingPathComponent:key] error:NULL] valueForKey:NSFileSize] unsignedLongLongValue];
            
            [[NSFileManager defaultManager] removeItemAtPath:[_cachePath stringByAppendingPathComponent:key] error:NULL];
            [[self imagesDictionary] removeObjectForKey:key];
            
            if(currentDiskCacheSize < MAX_IMAGE_CACHE_SIZE)
                break;
        }
    }
    
    [self writePlistToFile];
}

- (void)purgeImagesAtURLs:(NSArray *)URLs
{
    for(NSURL *aURL in URLs)
    {
        NSString *key = [self filePathForURL:aURL];
        [[NSFileManager defaultManager] removeItemAtPath:[_cachePath stringByAppendingPathComponent:key] error:NULL];
        [[self imagesDictionary] removeObjectForKey:key];
    }
    [self writePlistToFile];
}

#pragma mark -
#pragma mark Plist methods

- (void)updateImageInPlist:(NSString *)imagePath withLastModifiedDate:(NSString *)aDate
{
    [[self imagesDictionary] setValue:[NSDictionary dictionaryWithObjectsAndKeys:aDate, NLImageLastModifiedDate, [NSDate date], NLImageLastAccessedDate, nil] forKey:imagePath];
}

- (void)writePlistToFile
{
    [[self imagesDictionary] writeToFile:[self plistPath] atomically:YES];
}

- (NSString *)plistPath
{
    return [_cachePath stringByAppendingPathComponent:@"images.plist"];
}

@end

@implementation NLImageProviderHandler
@synthesize observer = _observer, continueLoading;

+ (id)handler; { return [[self alloc] init]; }

- (id)init
{
    if((self = [super init]) != nil)
        _loaders = [[NSMutableSet alloc] init];
    
    return self;
}

- (NSMutableSet *)loaders { return _loaders; }

@end
