//
//  NLHTTPRequestOperation.m
//  NLDataProvider
//
//  Created by Remy Demarest on 14/09/2010.
//  Copyright 2010 Nulayer Inc. All rights reserved.
//

#import "NLHTTPRequestOperation.h"
#import "NLDataProvider.h"
#import <NLFoundation/NLFoundation.h>

NSString *const NLDataProviderContentTypeOption     = @"contentType";
NSString *const NLDataProviderBodyOption            = @"body";
NSString *const NLDataProviderQueryParametersOption = @"query";
NSString *const NLHTTPRequestMethodOption           = @"HTTPMethod";
NSString *const NLDataProviderDataParserOption      = @"NLDataProviderDataParserOption";

@interface NLHTTPRequestOperation ()
{
    NSRunLoop *runloop;
    NSArray   *runloopModes;
}

@property(nonatomic, readwrite, copy) void (^requestCompletionBlock)(NSData *resources, NSHTTPURLResponse *response, NSError *error);

@property(nonatomic, readwrite, copy) NSURL        *requestURL;
@property(nonatomic, readwrite, copy) NSDictionary *requestOptions;

- (NSURLRequest *)NL_createRequest;
- (void)NL_connectionDidFinishWithError:(NSError *)error;
- (void)NL_returnResourcesWithError:(NSError *)error;
- (void)NL_cleanUp;

@end


@implementation NLHTTPRequestOperation
@synthesize requestCompletionBlock = _requestCompletionBlock, requestURL = _requestURL, requestOptions = _requestOptions, timeoutInterval = _timeoutInterval;

+ (id)requestOperationWithURL:(NSURL *)aURL options:(NSDictionary *)options completion:(void (^)(NSData *resources, NSHTTPURLResponse *response, NSError *error))completion;
{
    return [[self alloc] initWithURL:aURL options:options completion:completion];
}

- (id)init
{
    return nil;
}

- (id)initWithURL:(NSURL *)aURL options:(NSDictionary *)options completion:(void (^)(NSData *resources, NSHTTPURLResponse *response, NSError *error))completion;
{
    if((self = [super init]))
    {
        _timeoutInterval = 60.0;
        
        NSString *queryString = [[options objectForKey:NLDataProviderQueryParametersOption] queryString];
        _requestURL      = [([queryString length] > 0 ? [aURL URLByAppendingQueryString:queryString] : aURL) copy];
        _requestOptions  = [options copy];
        _connection      = [[NSURLConnection alloc] initWithRequest:[self NL_createRequest] delegate:self startImmediately:NO];
        
        runloop          = [NSRunLoop currentRunLoop];
        runloopModes     = [[NSArray alloc] initWithObjects:NSRunLoopCommonModes, [[NSRunLoop currentRunLoop] currentMode], nil];
        
        [_connection scheduleInRunLoop:runloop forMode:NSRunLoopCommonModes];
        
        NSString *mode = [[NSRunLoop currentRunLoop] currentMode];
        if(mode != nil) [_connection scheduleInRunLoop:runloop forMode:mode];
        
        [self setRequestCompletionBlock:completion];
    }
    return self;
}

- (void)dealloc
{
    [_connection cancel];
}

#pragma mark -
#pragma mark Accessors

- (BOOL)isExecuting  { return _requestFlags.isExecuting; }
- (BOOL)isFinished   { return _requestFlags.isFinished;  }
- (BOOL)isCancelled  { return _requestFlags.isCancelled; }
- (BOOL)isConcurrent { return YES;                       }

- (id<NLHTTPRequestOperationDelegate>)delegate; { return _delegate; }
- (void)setDelegate:(id<NLHTTPRequestOperationDelegate>)value
{
    @synchronized(self)
    {
        if(_delegate != value)
        {
            _delegate = value;
            
            _requestFlags.delegateDidBegin    = [_delegate respondsToSelector:@selector(requestOperationDidBegin:)];
            _requestFlags.delegateDidCancel   = [_delegate respondsToSelector:@selector(requestOperationDidCancel:)];
            _requestFlags.delegateDidDownload = [_delegate respondsToSelector:@selector(requestOperation:didDownloadContentOfSize:expectedSize:)];
            _requestFlags.delegateDidFinish   = [_delegate respondsToSelector:@selector(requestOperationDidFinish:)];
        }
    }
}

#pragma mark -
#pragma mark NSOperation methods

- (void)start
{
    @synchronized(self)
    {
        if(_requestFlags.isCancelled || _requestFlags.isFinished) return;
        
        [self willChangeValueForKey:@"isExecuting"];
        _requestFlags.isExecuting = YES;
        [self didChangeValueForKey:@"isExecuting"];
        
        if(_connection == nil) [self cancel];
        else
        {
            [_connection start];
            if(_requestFlags.delegateDidBegin) [runloop performSelector:@selector(requestOperationDidBegin:) target:[self delegate] argument:self order:0 modes:runloopModes];
        }
    }
}

- (void)cancel
{
    @synchronized(self)
    {
        [_connection cancel];
        RELEASE(_connection);
        
        [self willChangeValueForKey:@"isCancelled"];
        _requestFlags.isCancelled = YES;
        [self didChangeValueForKey:@"isCancelled"];
        
        if(_requestFlags.delegateDidCancel) [[self delegate] requestOperationDidCancel:self];
        [self setDelegate:nil];
        
        [self NL_connectionDidFinishWithError:nil];
    }
}

- (NSURLRequest *)NL_createRequest;
{
    NSMutableURLRequest *request = [[NSMutableURLRequest alloc] init];
    
    [self configureRequest:request];
    
    return [request copy];
}

- (void)configureRequest:(NSMutableURLRequest *)request
{
    NSDictionary *options = [self requestOptions];
    NSString     *method  = [options objectForKey:NLHTTPRequestMethodOption] ? : @"GET";
    
    [request setHTTPShouldUsePipelining:YES];
    [request setCachePolicy:NSURLRequestReloadIgnoringLocalAndRemoteCacheData];
    [request setTimeoutInterval:_timeoutInterval];
    [request setHTTPMethod:method];
    
    NSDictionary *headers = [options objectForKey:NLDataProviderHTTPHeadersOptions];
    if(headers != nil) [request setAllHTTPHeaderFields:headers];
    
    [request setURL:_requestURL];
    
    // FIXME: That should not be here...
    if([method isEqualToString:@"POST"] || [method isEqualToString:@"PUT"])
    {
        NSString *contentType = [options objectForKey:NLDataProviderContentTypeOption];
        
        if(contentType != nil)
            [request setValue:contentType forHTTPHeaderField:@"Content-Type"];
        
        id data = [options objectForKey:NLDataProviderBodyOption];
        
        if([data isKindOfClass:[NSData class]])
            [request setHTTPBody:data];
        else if([data isKindOfClass:[NSInputStream class]])
            [request setHTTPBodyStream:data];
    }
}

#pragma mark -
#pragma mark Cleaning up

- (void)NL_connectionDidFinishWithError:(NSError *)error;
{
    @synchronized(self)
    {
        if(_requestFlags.isExecuting)
        {
            [self willChangeValueForKey:@"isExecuting"];
            _requestFlags.isExecuting = NO;
            [self didChangeValueForKey:@"isExecuting"];
            
            [self willChangeValueForKey:@"isFinished"];
            _requestFlags.isFinished = YES;
            [self didChangeValueForKey:@"isFinished"];
        }
        
        [self NL_returnResourcesWithError:error];
        
        [self NL_cleanUp];
        
        if(!_requestFlags.isCancelled && _requestFlags.delegateDidFinish)
            [[self delegate] requestOperationDidFinish:self];
        
        [self setDelegate:nil];
    }
}

- (void)NL_returnResourcesWithError:(NSError *)error;
{
    @synchronized(self)
    {
        if([self isCancelled]) return;
        
        id returned = nil;
        
        if(error == nil && _response != nil)
        {
            NSInteger statusCode = [_response statusCode];
            
            if(!(200 <= statusCode && statusCode != 304 && statusCode < 400))
            {
                NSInteger     statusCode       = [_response statusCode];
                
                NSDictionary *headers          = [_response allHeaderFields];
                NSString     *errorReason      = [NSString stringWithFormat:@"%d Error: %@", statusCode, [NSHTTPURLResponse localizedStringForStatusCode:statusCode]];
                NSString     *errorDescription = [NSHTTPURLResponse localizedStringForStatusCode:statusCode];
                NSDictionary *userInfo         = [NSDictionary dictionaryWithObjectsAndKeys:
                                                  errorReason     , NSLocalizedFailureReasonErrorKey,
                                                  errorDescription, NSLocalizedDescriptionKey       ,
                                                  headers         , @"headers"                      ,
                                                  [NSNumber numberWithLongLong:[_response expectedContentLength]], @"expectedContentLength",
                                                  ([_response suggestedFilename] ? : @"No file name")     , @"suggestedFilename",
                                                  ([_response textEncodingName]  ? : @"No encoding name") , @"textEncodingName",
                                                  [NSNumber numberWithLongLong:[_responseData length]]    , @"data length",
                                                  ([_response URL] ? : (id) @"No URL !"), @"URL",
                                                  nil];
                error = [NSError errorWithDomain:NLDataProviderErrorDomain code:[_response statusCode] userInfo:userInfo];
            }
            returned = _responseData;
        }
        
        _requestCompletionBlock(returned, _response, error);
    }
}

- (void)NL_cleanUp;
{
    @synchronized(self)
    {
        RELEASE(_connection);
        RELEASE(_response);
        RELEASE(_responseData);
        RELEASE(_requestCompletionBlock);
    }
}

#pragma mark -
#pragma mark NSURLConnection

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSHTTPURLResponse *)response
{
    if(![response isKindOfClass:[NSHTTPURLResponse class]])
    {
        [_connection cancel];
        RELEASE(_connection);
        
        [self NL_connectionDidFinishWithError:[NSError errorWithDomain:NLDataProviderErrorDomain code:NLDataProviderInvalidResponseError userInfo:nil]];
        return;
    }
    
    _response = response;

    // Long cat is loooooooooong !
    long long cat = [response expectedContentLength];
    
    if(!_requestFlags.isCancelled && _requestFlags.delegateDidDownload)
        [[self delegate] requestOperation:self didDownloadContentOfSize:0 expectedSize:cat];
    
    if(cat == NSURLResponseUnknownLength || cat < 0)
        cat = 0;
    
    _responseData = [[NSMutableData alloc] initWithCapacity:cat];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
    [_responseData appendData:data];
    
    if(!_requestFlags.isCancelled && _requestFlags.delegateDidDownload)
        [[self delegate] requestOperation:self didDownloadContentOfSize:[_responseData length] expectedSize:[_response expectedContentLength]];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
    if(_connection == connection)
    {
        NSLog(@"error: %@, userInfo: %@", error, [error userInfo]);
        [self NL_connectionDidFinishWithError:error];
    }
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
    if(_connection == connection) [self NL_connectionDidFinishWithError:nil];
}

@end
