//
//  NLJSONDataParser.m
//  NLDataProvider
//
//  Created by Remy Demarest on 15/09/2010.
//  Copyright 2010 Nulayer Inc. All rights reserved.
//

#import "NLJSONDataParser.h"
#import <NLFoundation/NLFoundation.h>
#import "NLURLParser.h"
#import <Foundation/Foundation.h>

@interface NLJSONDataParser ()
- (NSStringEncoding)NL_encodingForOptions:(NSDictionary *)options;
- (BOOL)NL_isTypeValidForOptions:(NSDictionary *)options error:(NSError **)outError;
- (id)NL_JSONObjectForData:(NSData *)data options:(NSDictionary *)options error:(NSError **)outError;
@end


@implementation NLJSONDataParser

- (id)init
{
    if((self = [super init]))
        _URLParser = [[NLURLParser alloc] initWithCopyObjects:YES];
    
    return self;
}

- (NSStringEncoding)NL_encodingForOptions:(NSDictionary *)options;
{
    NSStringEncoding encoding = NSUTF8StringEncoding;
    
    id enc = [options objectForKey:NLDataParserNSStringEncodingOption];
    
    if(enc != nil) encoding = [enc integerValue];
    else
    {
        enc = [options objectForKey:NLDataParserIANACharSetNameOption];
        
        if(enc != nil)
        {
            CFStringEncoding cfEncoding = CFStringConvertIANACharSetNameToEncoding((__bridge CFStringRef)enc);
            encoding = CFStringConvertEncodingToNSStringEncoding(cfEncoding);
        }
    }
    
    return encoding;
}

- (BOOL)NL_isTypeValidForOptions:(NSDictionary *)options error:(NSError **)outError;
{
    NSString *MIMEType = [options objectForKey:NLDataParserMIMETypeOption];
    NSString *fileExt  = nil;
    
    if(MIMEType != nil && ![self canParseMIMEType:MIMEType]) goto error;
    else if(MIMEType == nil)
    {
        fileExt = [options objectForKey:NLDataParserSuggestedFilenameOption];
        
        if(fileExt != nil && ![self canParsePathExtension:fileExt]) goto error;
    }
    
    return YES;
    
error:
    
    if(outError != NULL)
    {
        *outError = [NSError errorWithDomain:NLDataParserErrorDomain
                                        code:NLDataParserErrorUnrecognizedTypeError
                                    userInfo:
                     [NSDictionary dictionaryWithObjectsAndKeys:
                      NSLocalizedString(@"Unrecognized Type", @"Parser type error description"), NSLocalizedDescriptionKey,
                      [NSString stringWithFormat:NSLocalizedString(@"The resource type is not supported. MIME Type: %@, File type %@", @"Parser type error reason"), MIMEType, fileExt], NSLocalizedFailureReasonErrorKey,
                      nil]];
    }
    
    return NO;
}

- (id)NL_JSONObjectForData:(NSData *)data options:(NSDictionary *)options error:(NSError **)outError;
{
    if([data length] == 0)
    {
        if(outError != NULL)
            *outError = nil;
        
        return nil;
    }
    
    NSError *parserError = nil;
    
    id results = nil;
    
    results = [NSJSONSerialization JSONObjectWithData:data options:0 error:&parserError];
    
    if(results == nil && outError != NULL)
        *outError = parserError;
    
    return results;
}

- (BOOL)canParseMIMEType:(NSString *)MIMEType;
{
    return ([MIMEType rangeOfString:@"json"].location != NSNotFound);
}

- (BOOL)canParsePathExtension:(NSString *)extension;
{
    return [extension isEqualToString:@"json"];
}

- (id)parsedObjectForData:(NSData *)data withURL:(NSURL *)aURL options:(NSDictionary *)options error:(NSError **)outError;
{
    if(![self NL_isTypeValidForOptions:options error:outError]) return nil;
    
    id resource = [self NL_JSONObjectForData:data options:options error:outError];
    
    if(resource == nil) return nil;
    
    id(^parser)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError) = [self parserBlockForURL:aURL];
    
    return (parser != nil ? parser(resource, aURL, options, outError) : resource);
}

#pragma mark -
#pragma mark Proxy methods for NLURLParser

- (id(^)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError))parserBlockForPath:(NSString *)aPath;
{
    return [_URLParser objectForPath:aPath];
}

- (id(^)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError))parserBlockForURL:(NSURL *)aURL;
{
    return [_URLParser objectForURL:aURL];
}

- (void)setParseScheme:(NSString *)aScheme withParserBlock:(id(^)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError))block;
{
    [_URLParser setObject:block forURLScheme:aScheme];
}

- (void)removeParserForScheme:(NSString *)aScheme;
{
    [_URLParser removeObjectForURLScheme:aScheme];
}

// The path accept wildcards as follow: */prefix suffix/* */infix/* 
- (void)setParsePath:(NSString *)aPath withParserBlock:(id(^)(id JSONData, NSURL *path, NSDictionary *options, NSError **outError))block;
{
    [_URLParser setObject:block forURLPath:aPath];
}

- (void)removeParserForPath:(NSString *)aPath;
{
    [_URLParser removeObjectForURLPath:aPath];
}

- (id (^)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError))defaultParser;
{
    return [_URLParser defaultObject];
}

- (void)setDefaultParser:(id (^)(id JSONData, NSURL *URL, NSDictionary *options, NSError **outError))value
{
    [_URLParser setDefaultObject:value];
}

@end
