//
//  NLDataProviderOperation.m
//  NLDataProvider
//
//  Created by Remy Demarest on 11/11/2010.
//  Copyright 2010 Nulayer Inc. All rights reserved.
//

#define HC_SHORTHAND

#import <Foundation/Foundation.h>
#import <NLFoundation/NLFoundation.h>
#import <Cedar/Cedar.h>
#import <OCHamcrest/OCHamcrest.h>

#import "NLReplaceClass.h"
#import "NLHTTPRequestOperation.h"
#import "FAKE_NSHTTPURLResponse.h"
#import "FAKE_NSURLConnection.h"

@interface NLHTTPRequestOperationTest : CDRSpec @end

@implementation NLHTTPRequestOperationTest

- (void)declareBehaviors
{
    describe(@"Running a GET request", ^{
        
        beforeEach(^{
            NLReplaceClass([NSURLConnection class], [FAKE_NSURLConnection class]);
        });
        
        afterEach(^{
            NLRestoreClass([NSURLConnection class]);
        });
        
        __block NSData                 *resourcesObject = nil;
        __block NSHTTPURLResponse      *responseObject  = nil;
        __block NSError                *errorObject     = nil;
        __block NLHTTPRequestOperation *operation       = nil;
        
        afterEach(^{
            RELEASE(operation      );
            RELEASE(responseObject );
            RELEASE(resourcesObject);
            RELEASE(errorObject);
        });
        
        describe(@"that succeeds", ^{
            beforeEach(^{
                operation =
                [[NLHTTPRequestOperation alloc] initWithURL:[NSURL URLWithString:NLOperationGETRequest] options:nil completion:
                 ^(NSData *resources, NSHTTPURLResponse *response, NSError *error)
                 {
                     resourcesObject = [resources retain];
                     responseObject  = [response  retain];
                     errorObject     = [error     retain];
                 }];
                
                [operation start];
                [[NSRunLoop currentRunLoop] run];
            });
            
            it(@"should return the response data", ^{
                assertThat(resourcesObject, is(equalTo([NLOperationGETRequest dataUsingEncoding:NSUTF8StringEncoding])));
            });
            
            it(@"should return a response with statusCode 200", ^{
                assertThatInteger([responseObject statusCode], is(equalToInteger(200)));
            });
            
            it(@"should return a nil error object", ^{
                assertThat(errorObject, is(equalTo(nil)));
            });
            
            describe(@"operation", ^{
                it(@"should not be cancelled", ^{
                    assertThatBool([operation isCancelled], is(equalToBool(NO)));
                });
                
                it(@"should be finished", ^{
                    assertThatBool([operation isFinished] , is(equalToBool(YES)));
                });
                
                it(@"should not be executing", ^{
                    assertThatBool([operation isExecuting], is(equalToBool(NO)));
                });
            });
        });
        
        describe(@"that returns an error", ^{
            beforeEach(^{
                operation =
                [[NLHTTPRequestOperation alloc] initWithURL:[NSURL URLWithString:NLOperationGETError] options:nil completion:
                 ^(NSData *resources, NSHTTPURLResponse *response, NSError *error)
                 {
                     resourcesObject = [resources retain];
                     responseObject  = [response  retain];
                     errorObject     = [error     retain];
                 }];
                
                [operation start];
                [[NSRunLoop currentRunLoop] run];
            });
            
            it(@"should return a nil response object", ^{
                assertThat(responseObject, is(equalTo(nil)));
            });
            
            it(@"should return a nil resources object", ^{
                assertThat(resourcesObject, is(equalTo(nil)));
            });
            
            it(@"should return an error object with a non-zero code", ^{
                assertThatInteger([errorObject code], isNot(equalToInteger(0)));
            });
            
            describe(@"operation", ^{
                it(@"should not be cancelled", ^{
                    assertThatBool([operation isCancelled], is(equalToBool(NO)));
                });
                
                it(@"should be finished", ^{
                    assertThatBool([operation isFinished] , is(equalToBool(YES)));
                });
                
                it(@"should not be executing", ^{
                    assertThatBool([operation isExecuting], is(equalToBool(NO)));
                });
            });
        });
        
         NSString *expected = @"The block should not be run";
        __block NSString *result = nil;
        
        describe(@"that is cancelled before completing", ^{
            beforeEach(^{
                result = expected;
                operation =
                [[NLHTTPRequestOperation alloc] initWithURL:[NSURL URLWithString:NLOperationGETError] options:nil completion:
                 ^(NSData *resources, NSHTTPURLResponse *response, NSError *error)
                 {
                     result = @"The block was run...";
                 }];
                
                [operation start];
                
                [operation cancel];
                
                [[NSRunLoop currentRunLoop] run];
            });
            
            it(@"should not run the completion block", ^{
                assertThat(result, is(equalTo(expected)));
            });
            
            describe(@"operation", ^{
                it(@"should be cancelled", ^{
                    assertThatBool([operation isCancelled], is(equalToBool(YES)));
                });
                
                it(@"should be finished", ^{
                    assertThatBool([operation isFinished] , is(equalToBool(YES)));
                });
                
                it(@"should not be executing", ^{
                    assertThatBool([operation isExecuting], is(equalToBool(NO)));
                });
            });
        });
        
        describe(@"with a connection that cannot run", ^{
            beforeEach(^{
                operation =
                [[NLHTTPRequestOperation alloc] initWithURL:[NSURL URLWithString:NLOperationGETDie] options:nil completion:
                 ^(NSData *resources, NSHTTPURLResponse *response, NSError *error)
                 {
                     result = @"The block was run...";
                 }];
                
                [operation start];
                
                [operation cancel];
                
                [[NSRunLoop currentRunLoop] run];
            });
            
            it(@"should not run the completion block", ^{
                assertThat(result, is(equalTo(expected)));
            });
            
            describe(@"operation", ^{
                it(@"should be cancelled", ^{
                    assertThatBool([operation isCancelled], is(equalToBool(YES)));
                });
                
                it(@"should be finished", ^{
                    assertThatBool([operation isFinished] , is(equalToBool(YES)));
                });
                
                it(@"should not be executing", ^{
                    assertThatBool([operation isExecuting], is(equalToBool(NO)));
                });
            });
        });
    });
    
    describe(@"Running a POST request", ^{
        
    });
    
    describe(@"Running a PUT request", ^{
        
    });
    
    describe(@"Running a DELETE request", ^{
        
    });
}

@end
