//
//  NLLocalStorageProviderTests.m
//  NLDataProvider
//
//  Created by Remy Demarest on 08/11/2010.
//  Copyright 2010 Nulayer Inc. All rights reserved.
//

#define HC_SHORTHAND

#import <Foundation/Foundation.h>
#import <NLFoundation/NLFoundation.h>
#import <Cedar/Cedar.h>
#import <OCHamcrest/OCHamcrest.h>

#import "NLLocalStorageDataProvider.h"
#import "NLFileMetadata.h"

typedef enum NLDataProviderAction
{
    NLDataProviderActionReturnData,
    NLDataProviderActionReturn304,
    NLDataProviderActionReturnNewData,
} NLDataProviderAction;

@interface NLDataProviderDelegate : NSObject <NLLocalStorageDataProviderDelegate, NLDataProvider>
{
    NLDataProviderAction currentAction;
}
@property(nonatomic) NLDataProviderAction currentAction;
@end

static NSString *const NLFallbackWhateverResult      = @"NLFallbackWhateverResult";
static NSString *const NLCacheFallbackWhateverResult = @"NLCacheFallbackWhateverResult";
static NSString *const NLFallbackNewDataCacheResult  = @"NLFallbackNewDataCacheResult";


@interface NLLocalStorageDataProviderTest : CDRSpec @end

@implementation NLLocalStorageDataProviderTest

- (void)declareBehaviors
{
    NSString *path = [NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES) lastObject];
    [[NSFileManager defaultManager] removeItemAtPath:path error:NULL];
    
    __block NLDataProviderDelegate     *testDelegate = [[NLDataProviderDelegate alloc] init];
    __block NLLocalStorageDataProvider *dataProvider = [[NLLocalStorageDataProvider alloc] initWithBasePath:path];
    
    
    beforeEach(^{
        testDelegate = [[NLDataProviderDelegate alloc] init];
        dataProvider = [[NLLocalStorageDataProvider alloc] initWithBasePath:path];
        
        [dataProvider setDelegate:testDelegate];
        [dataProvider setFallbackProvider:testDelegate];
    });
    
    __block id       observerObject  = nil;
    __block id       resourcesObject = nil;
    __block NSError *errorObject     = nil;
    
    afterEach(^{
        [[NSFileManager defaultManager] removeItemAtPath:path error:NULL];
        
        RELEASE(observerObject);
        RELEASE(resourcesObject);
        RELEASE(errorObject);
        
        RELEASE(testDelegate);
        RELEASE(dataProvider);
    });
    
    describe(@"Request resources that do not exist", ^{
        beforeEach(^{
            observerObject =
            [[dataProvider loadResourcesForURL:[NSURL URLWithString:@"about:whatever"] options:nil completion:
              ^(id resources, NSDictionary *extraInfo, NSError *error)
              {
                  resourcesObject = [resources retain];
                  errorObject     = [error     retain];
              }] retain];
        });
        
        it(@"should return a nil observer", ^{
            assertThat(observerObject, is(equalTo(nil)));
        });
        
        it(@"should return a nil resources object", ^{
            assertThat(resourcesObject, is(equalTo(nil)));
        });
        
        it(@"should return an NLDataProviderContentUnavailableError error", ^{
            assertThatInt([errorObject code], equalToInt(NLDataProviderContentUnavailableError));
        });
    });
    
    describe(@"Request handled by the fallback data provider", ^{
        beforeEach(^{
            observerObject =
            [[dataProvider loadResourcesForURL:[NSURL URLWithString:@"fallback:whatever"] options:nil completion:
              ^(id resources, NSDictionary *extraInfo, NSError *error)
              {
                  resourcesObject = [resources retain];
                  errorObject     = [error     retain];
              }] retain];
        });
        
        it(@"should return an observer object", ^{
            assertThat(observerObject, is(equalTo(NLFallbackWhateverResult)));
        });
        
        it(@"should return fallback data object", ^{
            assertThat(resourcesObject, is(equalTo(NLFallbackWhateverResult)));
        });
        
        it(@"should not return an error object", ^{
            assertThat(errorObject, is(equalTo(nil)));
        });
    });
    
    describe(@"Content caching", ^{
        NSURL    *resourceURL = [NSURL URLWithString:@"cache:whatever"];
        NSString *localPath   = [[path stringByAppendingPathComponent:@"Storage"] stringByAppendingPathComponent:@"whatever"];
        
        it(@"should generate the file path for the resourceURL", ^{
            assertThat([dataProvider resourcePathForRequestURL:resourceURL], equalTo(localPath));
        });
        
        void (^runGetRequestWithAction)(NLDataProviderAction, BOOL) =
        ^ (NLDataProviderAction action, BOOL saveLocals)
        {
            [testDelegate setCurrentAction:action];
            
            id obs =
            [dataProvider loadResourcesForURL:resourceURL options:nil completion:
             ^(id resources, NSDictionary *extraInfo, NSError *error)
             {
                 if(saveLocals)
                 {
                     resourcesObject = [[NSString alloc] initWithData:resources encoding:NSUTF8StringEncoding];
                     errorObject     = [error     retain];
                 }
             }];
            
            if(saveLocals) observerObject = [obs retain];
        };
        
        describe(@"when the data are not yet cached to the disk", ^{
            beforeEach(^{
                runGetRequestWithAction(NLDataProviderActionReturnData, YES);
            });
            
            it(@"should return the observer object of the fallback data provider", ^{
                assertThat(observerObject, is(equalTo(NLCacheFallbackWhateverResult)));
            });
            
            it(@"should return the resources of the fallback data provider", ^{
                assertThat(resourcesObject, is(equalTo(NLFallbackWhateverResult)));
            });
            
            it(@"should not return an error object", ^{
                assertThat(errorObject, is(equalTo(nil)));
            });
            
            it(@"should save the resources to a file on the local storage", ^{
                assertThatBool([[NSFileManager defaultManager] fileExistsAtPath:localPath], equalToBool(YES));
            });
        });
        
        describe(@"when the data are already cached to the disk", ^{
            beforeEach(^{
                runGetRequestWithAction(NLDataProviderActionReturnData, NO);
                runGetRequestWithAction(NLDataProviderActionReturnData, YES);
            });
            
            it(@"should return a nil observer object", ^{
                assertThat(observerObject, is(equalTo(nil)));
            });
            
            it(@"should return the data saved on the disk", ^{
                assertThat(resourcesObject, is(equalTo(NLFallbackWhateverResult)));
            });
            
            it(@"should not return an error object", ^{
                assertThat(errorObject, is(equalTo(nil)));
            });
        });
        
        describe(@"when the data on the disk are outdated but not on the server", ^{
            beforeEach(^{
                runGetRequestWithAction(NLDataProviderActionReturnData, NO);
                
                [NSThread sleepForTimeInterval:0.2];
                
                [[dataProvider fileMetadataForResourceURL:resourceURL] setLastAccessDate:[NSDate distantPast]];
                runGetRequestWithAction(NLDataProviderActionReturn304, YES);
            });
            
            it(@"should return the observer block of the fallback data provider", ^{
                assertThat(observerObject, is(equalTo(NLCacheFallbackWhateverResult)));
            });
            
            it(@"should return the data from the local storage cache", ^{
                assertThat(resourcesObject, is(equalTo(NLFallbackWhateverResult)));
            });
            
            it(@"should not return an error object", ^{
                assertThat(errorObject, is(equalTo(nil)));
            });
        });
        
        describe(@"when the data on the disk and the server are outdated", ^{
            beforeEach(^{
                runGetRequestWithAction(NLDataProviderActionReturnData, NO);
                
                [NSThread sleepForTimeInterval:0.2];
                
                [[dataProvider fileMetadataForResourceURL:resourceURL] setLastAccessDate:[NSDate distantPast]];
                runGetRequestWithAction(NLDataProviderActionReturnNewData, YES);
            });
            
            it(@"should return the observer block of the fallback data provider", ^{
                assertThat(observerObject, is(equalTo(NLFallbackNewDataCacheResult)));
            });
            
            it(@"should return the new data from the server", ^{
                assertThat(resourcesObject, is(equalTo(NLFallbackNewDataCacheResult)));
            });
            
            it(@"should not return an error object", ^{
                assertThat(errorObject, is(equalTo(nil)));
            });
        });
    });
}
@end

@implementation NLDataProviderDelegate
@synthesize currentAction;

- (NSURL *)baseURL; { return nil; }
- (NSURL *)requestURLForResourceURL:(NSURL *)resourceURL; { return nil; }

- (BOOL)localStorageDataProvider:(NLLocalStorageDataProvider *)sender shouldHandleRequestForResourceURL:(NSURL *)resourceURL;
{
    return ![[resourceURL scheme] isEqualToString:@"fallback"];
}

- (id)loadResourcesForURL:(NSURL *)resourceURL options:(NSDictionary *)options completion:(void(^)(id resources, NSDictionary *extraInfo, NSError *error))completion;
{
    if([[resourceURL scheme] isEqualToString:@"fallback"])
    {
        completion(NLFallbackWhateverResult, nil, nil);
        return NLFallbackWhateverResult;
    }
    else if([[resourceURL scheme] isEqualToString:@"cache"])
    {
        switch(currentAction)
        {
            case NLDataProviderActionReturnData    :
                completion([NLFallbackWhateverResult dataUsingEncoding:NSUTF8StringEncoding], [NSDictionary dictionaryWithObject:[NSDate dateWithTimeIntervalSinceNow:0.2] forKey:NLDataProviderReloadAfterDateExtraInfo], nil);
                return NLCacheFallbackWhateverResult;
                
            case NLDataProviderActionReturn304     :
                completion(nil, nil, [NSError errorWithDomain:@"test domain" code:NLDataProviderContentNotModifiedError userInfo:nil]);
                return NLCacheFallbackWhateverResult;
                
            case NLDataProviderActionReturnNewData :
                completion([NLFallbackNewDataCacheResult dataUsingEncoding:NSUTF8StringEncoding], nil, nil);
                return NLFallbackNewDataCacheResult;
        }
    }
    else if(completion != nil)
    {
        completion(nil, nil, nil);
    }
    
    return nil;
}

// TODO: Implement these methods
//- (id)createResource:(id)resources forURL:(NSURL *)resourceURL options:(NSDictionary *)options completion:(void(^)(id resources, NSDictionary *extraInfo, NSError *error))completion;
//- (id)updateResource:(id)resources forURL:(NSURL *)resourceURL options:(NSDictionary *)options completion:(void(^)(id resources, NSDictionary *extraInfo, NSError *error))completion;
//- (id)deleteResourceForURL:(NSURL *)resourceURL options:(NSDictionary *)options completion:(void(^)(id resources, NSDictionary *extraInfo, NSError *error))completion;

// Observer is the opaque object returned by -loadResourcesForURL:options:completion:
- (void)removeCompletionBlock:(id)observer;
{
    
}

@end

