//
//  NLRemoteDataProviderTest.m
//  NLDataProvider
//
//  Created by Remy Demarest on 10/11/2010.
//  Copyright 2010 Nulayer Inc. All rights reserved.
//


#define HC_SHORTHAND

#import <Foundation/Foundation.h>
#import <NLFoundation/NLFoundation.h>
#import <Cedar/Cedar.h>
#import <OCHamcrest/OCHamcrest.h>

#import "NLReplaceClass.h"
#import "FAKE_NSHTTPURLResponse.h"
#import "FAKE_NLHTTPRequestOperation.h"
#import "NLHTTPRequestOperation.h"
#import "NLRemoteDataProvider.h"
#import "NLOperationQueue.h"

#define MAKE_CONTEXT(ctx) ^ NSDictionary * { return ctx; }

@interface NLRemoteDataProviderDelegate : NSObject <NLRemoteDataProviderDelegate, NLDataParser>
{
    NSCountedSet *shouldBeginRequests;
    NSCountedSet *beganRequests;
    NSCountedSet *cancelledRequests;
    NSCountedSet *finishedRequests;
    NSCountedSet *returningRequests;
}
@property(nonatomic, readonly) NSCountedSet *shouldBeginRequests;
@property(nonatomic, readonly) NSCountedSet *beganRequests;
@property(nonatomic, readonly) NSCountedSet *cancelledRequests;
@property(nonatomic, readonly) NSCountedSet *finishedRequests;
@property(nonatomic, readonly) NSCountedSet *returningRequests;

@property(nonatomic) BOOL rewriteApproved;
@property(nonatomic) BOOL didEndAllReceived;
@end

SHARED_EXAMPLE_GROUPS_BEGIN(Specs)
/*
NLOperationGETRewriteURL;
NLOperationGETRewrittenURL;
NLOperationGETDelegatePrevent;
NLOperationGETParsedMIMEType;
NLOperationGETParsedURLExt;
NLOperationGETParsedSuggest;
NLOperationGETRawData;
NLOperationGETError;
NLOperationGETCancel;
 */

sharedExamplesFor(@"a successeful request with raw data", ^(NSDictionary *(^context)(void)) {
    it(@"should return a non-nil observer", ^{
        assertThat([context() objectForKey:@"observer"], isNot(equalTo([NSNull null])));
    });
    
    it(@"should return nil as the error value", ^{
        assertThat([context() objectForKey:@"error"], is(equalTo([NSNull null])));
    });
    
    it(@"should return resources of type NSData", ^{
        assertThat([context() objectForKey:@"resources"], instanceOf([NSData class]));
    });
});

sharedExamplesFor(@"a request that does not return", ^(NSDictionary *(^context)(void)) {
    it(@"should return a non-nil observer", ^{
        assertThat([context() objectForKey:@"observer"], isNot(equalTo([NSNull null])));
    });
    
    it(@"should not execute the completion block", ^{
        assertThat(nil, allOf(equalTo([context() objectForKey:@"error"]), equalTo([context() objectForKey:@"resources"]), nil));
    });
});

sharedExamplesFor(@"a request that does not start", ^(NSDictionary *(^context)(void)) {
    it(@"should return a nil observer", ^{
        assertThat([context() objectForKey:@"observer"], is(equalTo([NSNull null])));
    });
    
    it(@"should not execute the completion block", ^{
        assertThat(nil, allOf(equalTo([context() objectForKey:@"error"]), equalTo([context() objectForKey:@"resources"]), nil));
    });
});

sharedExamplesFor(@"NLOperationGETRawData request", ^(NSDictionary *(^context)(void)) {
    itShouldBehaveLikeWithContext(nil, @"a successeful request with raw data", context);
    
    it(@"should return resources equal to NLOperationGETRawData encoded string", ^{
        assertThat([context() objectForKey:@"resources"], is(equalTo([NLOperationGETRawData dataUsingEncoding:NSUTF8StringEncoding])));
    });
});

sharedExamplesFor(@"NLOperationGETRewriteURL request", ^(NSDictionary *(^context)(void)) {
    itShouldBehaveLikeWithContext(nil, @"a successeful request with raw data", context);
    
    it(@"should return resources equal to NLOperationGETRewrittenURL encoded string", ^{
        assertThat([context() objectForKey:@"resources"], is(equalTo([NLOperationGETRewrittenURL dataUsingEncoding:NSUTF8StringEncoding])));
    });
});

sharedExamplesFor(@"NLOperationGETError request", ^(NSDictionary *(^context)(void)) {
    it(@"should return nil as the resources value", ^{
        assertThat([context() objectForKey:@"resources"], is(equalTo([NSNull null])));
    });
    
    describe(@"error", ^{
        it(@"should not be nil", ^{
            assertThat([context() objectForKey:@"error"], isNot(equalTo(nil)));
        });
        
        it(@"code should be equal to -1001", ^{
            assertThatInteger([[context() objectForKey:@"error"] code], is(equalToInteger(-1001)));
        });
    });
});

SHARED_EXAMPLE_GROUPS_END

@interface NLRemoteDataProviderTest : CDRSpec @end

@implementation NLRemoteDataProviderTest

- (void)declareBehaviors
{
    __block NLRemoteDataProviderDelegate *delegate     = nil;
    __block NLRemoteDataProvider         *dataProvider = nil;
    
    id (^loadRequestWithURL)(NSURL *, NSMutableDictionary *context) =
    ^ id (NSURL *url, NSMutableDictionary *context)
    {
        id observer =
        [dataProvider loadResourcesForURL:url options:nil completion:
         ^(id resources, NSDictionary *extraInfo, NSError *error)
         {
             [context setObject:(id)(error     ? : [NSNull null]) forKey:@"error"];
             [context setObject:(id)(resources ? : [NSNull null]) forKey:@"resources"];
         }];
        
        [context setObject:(id)(observer ? : [NSNull null]) forKey:@"observer"];
        
        return observer;
    };
    
    beforeEach(^{
        NLReplaceClass([NLHTTPRequestOperation class], [FAKE_NLHTTPRequestOperation class]);
        
        
        NLReplaceClass([NSOperationQueue class], [NLOperationQueue class]);
        
        dataProvider = [[NLRemoteDataProvider alloc] init];
        [dataProvider setBaseURL:[NSURL URLWithString:@"http://get.example.com/"]];
        
        NLRestoreClass([NSOperationQueue class]);
        
        // Delegate
        delegate = [[NLRemoteDataProviderDelegate alloc] init];
        [dataProvider setDelegate:delegate];
        [dataProvider setDataParsers:[NSArray arrayWithObject:delegate]];
    });
    
    afterEach(^{
        NLRestoreClass([NLHTTPRequestOperation class]);
        
        // Remove and reset the delegate
        [dataProvider setDataParsers:nil];
        [dataProvider setDelegate:nil];
        [delegate release];
        delegate = nil;
        
        [dataProvider release];
    });
    
    describe(@"Send a single request", ^{
        
        __block NSMutableDictionary *context = nil;
        
        void (^runRequestWithURL)(NSURL *) =
        ^(NSURL *url)
        {
            loadRequestWithURL(url, context);
            CFRunLoopRunInMode(kCFRunLoopDefaultMode, 10.0, NO);
        };
        
        beforeEach(^{ context = [[NSMutableDictionary alloc] init]; });
        
        afterEach(^{ [context release], context = nil; });
        
        describe(@"that has succeeded", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETRawData];
            
            beforeEach(^{ runRequestWithURL(url); });
            
            itShouldBehaveLikeWithContext(nil, @"NLOperationGETRawData request", ^{ return context; });
            
            describe(@"delegate", ^{
                it(@"should be asked whether to start the request", ^{
                    assertThat([delegate shouldBeginRequests], hasItem(url));
                });
                
                it(@"should be notified the request has started", ^{
                    assertThat([delegate beganRequests], hasItem(url));
                });
                
                it(@"should not be notified the request was cancelled", ^{
                    assertThat([delegate cancelledRequests], isNot(hasItem(url)));
                });
                
                it(@"should be notified data were returned", ^{
                    assertThat([delegate returningRequests]  , hasItem(url));
                });
                
                it(@"should be notified all requests are done", ^{
                    assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                });
            });
        });
        
        describe(@"that needs rewriting", ^{
            NSURL *url          = [NSURL URLWithString:NLOperationGETRewriteURL];
            NSURL *rewrittenURL = [NSURL URLWithString:NLOperationGETRewrittenURL];
            
            beforeEach(^{ runRequestWithURL(url); });
            
            itShouldBehaveLikeWithContext(nil, @"NLOperationGETRewriteURL request", MAKE_CONTEXT(context));
            
            describe(@"delegate", ^{
                it(@"should be asked whether to start the request for the original URL", ^{
                    assertThat([delegate shouldBeginRequests], hasItem(url));
                });
                
                it(@"should not be notified a request with the original URL has been started", ^{
                    assertThat([delegate beganRequests], isNot(hasItem(url)));
                });
                
                it(@"should be notified a request with the rewritten URL has been started", ^{
                    assertThat([delegate beganRequests], hasItem(rewrittenURL));
                });
                
                it(@"should not be notified a request with the original URL has finished", ^{
                    assertThat([delegate finishedRequests], isNot(hasItem(url)));
                });
                
                it(@"should be notified a request with the rewritten URL has finished", ^{
                    assertThat([delegate finishedRequests], hasItem(rewrittenURL));
                });
                
                it(@"should be notified a request with the original URL returns its resources", ^{
                    assertThat([delegate returningRequests], hasItem(url));
                });
                
                it(@"should not be notified a request with the rewritten URL returns its resources", ^{
                    assertThat([delegate returningRequests], isNot(hasItem(rewrittenURL)));
                });
                
                it(@"should have approved the URL rewritting", ^{
                    assertThatBool([delegate rewriteApproved], is(equalToBool(YES)));
                });
            });
        });
        
        describe(@"that is rejected by the delegate", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETDelegatePrevent];
            
            beforeEach(^{ runRequestWithURL(url); });
            
            itShouldBehaveLikeWithContext(nil, @"a request that does not start", ^{ return context; });
            
            describe(@"delegate", ^{
                it(@"should be asked whether to start the request", ^{
                    assertThat([delegate shouldBeginRequests], hasItem(url));
                });
                
                it(@"should not be notified the request has been cancelled", ^{
                    assertThat([delegate cancelledRequests], isNot(hasItem(url)));
                });
                
                it(@"should not be notified the request has been started", ^{
                    assertThat([delegate beganRequests], isNot(hasItem(url)));
                });
                
                it(@"should not be notified the request has finished", ^{
                    assertThat([delegate finishedRequests], isNot(hasItem(url)));
                });
                
                it(@"should not be notified the request has returned data", ^{
                    assertThat([delegate returningRequests], isNot(hasItem(url)));
                });
                
                it(@"should not be notified all request finished", ^{
                    assertThatBool([delegate didEndAllReceived], is(equalToBool(NO)));
                });
            });
        });
        
        describe(@"that is cancelled by the observer", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETCancel];
            
            beforeEach(^{
                id observer = loadRequestWithURL(url, context);
                
                CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 0.0, YES);
                [dataProvider removeCompletionBlock:observer];
                CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
            });
            
            itShouldBehaveLikeWithContext(nil, @"a request that does not return", ^{ return context; });
            
            describe(@"delegate", ^{
                it(@"should be asked whether to start the request", ^{
                    assertThat([delegate shouldBeginRequests], hasItem(url));
                });
                
                it(@"should be notified the request has started", ^{
                    assertThat([delegate beganRequests], hasItem(url));
                });
                
                it(@"should be notified the request was cancelled", ^{
                    assertThat([delegate cancelledRequests], hasItem(url));
                });
                
                it(@"should not be notified the request has finished", ^{
                    assertThat([delegate finishedRequests], isNot(hasItem(url)));
                });
                
                it(@"should not be notified the request returned data", ^{
                    assertThat([delegate returningRequests], isNot(hasItem(url)));
                });
                
                it(@"should be notified all request did end", ^{
                    assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                });
            });
        });
        
        describe(@"that returns an error", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETError];
            
            beforeEach(^{ runRequestWithURL(url); });
            
            itShouldBehaveLikeWithContext(nil, @"NLOperationGETError request", ^{ return context; });
            
            describe(@"delegate", ^{
                it(@"should not be notified the request has been cancelled", ^{
                    assertThat([delegate cancelledRequests], isNot(hasItem(url)));
                });
                
                it(@"should be notified the request did finish", ^{
                    assertThat([delegate finishedRequests], hasItem(url));
                });
                
                it(@"should be notified the request returned resources ", ^{
                    assertThat([delegate returningRequests], hasItem(url));
                });
            });
        });
    });
        
    describe(@"Send a multiple requests", ^{
        __block NSMutableDictionary *firstRequest, *secondRequest;
        
        beforeEach(^{
            firstRequest  = [[NSMutableDictionary alloc] init];
            secondRequest = [[NSMutableDictionary alloc] init];
        });
        
        afterEach(^{
            [firstRequest  release], firstRequest  = nil;
            [secondRequest release], secondRequest = nil;
        });
        
        describe(@"that load resources for the same URL", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETRawData];
            
            describe(@"and finishing successfully", ^{
                beforeEach(^{
                    loadRequestWithURL(url, firstRequest);
                    CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0.0, YES);
                    loadRequestWithURL(url, secondRequest);
                    CFRunLoopRunInMode(kCFRunLoopDefaultMode, 10.0, NO);
                });
                
                itShouldBehaveLikeWithContext(@"firstRequest", @"NLOperationGETRawData request", MAKE_CONTEXT(firstRequest));
                
                itShouldBehaveLikeWithContext(@"secondRequest", @"NLOperationGETRawData request", MAKE_CONTEXT(secondRequest));
                
                describe(@"delegate", ^{
                    it(@"should be asked whether to start the request twice", ^{
                        assertThatUnsignedInteger([[delegate shouldBeginRequests] countForObject:url], is(equalToUnsignedInteger(2)));
                    });
                    
                    it(@"should receive only one request notification for the URL", ^{
                        assertThatUnsignedInteger([[delegate beganRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                    });
                    
                    it(@"should not be notified the request was cancelled", ^{
                        assertThatUnsignedInteger([[delegate cancelledRequests] countForObject:url], is(equalToUnsignedInteger(0)));
                    });
                    
                    it(@"should be notified the request was finished once", ^{
                        assertThatUnsignedInteger([[delegate  finishedRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                    });
                    
                    it(@"should be notified data were returned by the request twice", ^{
                        assertThatUnsignedInteger([[delegate returningRequests] countForObject:url], is(equalToUnsignedInteger(2)));
                    });
                    
                    it(@"should be notified all requests did end", ^{
                        assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                    });
                });
            });
            
            describe(@"when one observer is removed", ^{
                beforeEach(^{
                    id observer = loadRequestWithURL(url, firstRequest);
                    CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0.0, YES);
                    
                    loadRequestWithURL(url, secondRequest);
                    
                    [dataProvider removeCompletionBlock:observer];
                    
                    CFRunLoopRunInMode(kCFRunLoopDefaultMode, 10.0, NO);
                });
                
                itShouldBehaveLikeWithContext(@"firstRequest", @"a request that does not return", MAKE_CONTEXT(firstRequest));
                
                itShouldBehaveLikeWithContext(@"secondRequest", @"NLOperationGETRawData request", MAKE_CONTEXT(secondRequest));
                
                it(@"should be asked whether to start the request twice", ^{
                    assertThatUnsignedInteger([[delegate shouldBeginRequests] countForObject:url], is(equalToUnsignedInteger(2)));
                });
                
                it(@"should receive only one request notification for the URL", ^{
                    assertThatUnsignedInteger([[delegate beganRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                });
                
                it(@"should not be notified the request was cancelled", ^{
                    assertThatUnsignedInteger([[delegate cancelledRequests] countForObject:url], is(equalToUnsignedInteger(0)));
                });
                
                it(@"should be notified the request was finished once", ^{
                    assertThatUnsignedInteger([[delegate  finishedRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                });
                
                it(@"should be notified data were returned by the request once", ^{
                    assertThatUnsignedInteger([[delegate returningRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                });
                
                it(@"should be notified all requests did end", ^{
                    assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                });
            });
            
            describe(@"when all observers are removed", ^{
                beforeEach(^{
                    id observer1 = loadRequestWithURL(url, firstRequest);
                    CFRunLoopRunInMode(kCFRunLoopDefaultMode, 0.0, YES);
                    
                    id observer2 = loadRequestWithURL(url, secondRequest);
                    
                    [dataProvider removeCompletionBlock:observer1];
                    [dataProvider removeCompletionBlock:observer2];
                    
                    CFRunLoopRunInMode(kCFRunLoopDefaultMode, 10.0, NO);
                });
                
                itShouldBehaveLikeWithContext(@"firstRequest", @"a request that does not return", MAKE_CONTEXT(firstRequest));
                
                itShouldBehaveLikeWithContext(@"secondRequest", @"a request that does not return", MAKE_CONTEXT(secondRequest));
                
                it(@"should be asked whether to start the request twice", ^{
                    assertThatUnsignedInteger([[delegate shouldBeginRequests] countForObject:url], is(equalToUnsignedInteger(2)));
                });
                
                it(@"should be notified the request started once", ^{
                    assertThatUnsignedInteger([[delegate beganRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                });
                
                it(@"should be notified the request was cancelled once", ^{
                    assertThatUnsignedInteger([[delegate cancelledRequests] countForObject:url], is(equalToUnsignedInteger(1)));
                });
                
                it(@"should not be notified the request was finished", ^{
                    assertThatUnsignedInteger([[delegate  finishedRequests] countForObject:url], is(equalToUnsignedInteger(0)));
                });
                
                it(@"should not be notified data were returned by the request", ^{
                    assertThatUnsignedInteger([[delegate returningRequests] countForObject:url], is(equalToUnsignedInteger(0)));
                });
                
                it(@"should be notified all requests did end", ^{
                    assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                });
            });
            
        });
        
        describe(@"that load resources with different request URLs", ^{
            describe(@"when a rewritten URL end up loading the same resource", ^{
                NSURL *url       = [NSURL URLWithString:NLOperationGETRewriteURL];
                NSURL *rewritten = [NSURL URLWithString:NLOperationGETRewrittenURL];
                
                beforeEach(^{
                    loadRequestWithURL(url, firstRequest);
                    CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 0.0, YES);
                    loadRequestWithURL(rewritten, secondRequest);
                    CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
                });
                
                itShouldBehaveLikeWithContext(@"request being rewrote", @"NLOperationGETRewriteURL request", MAKE_CONTEXT(firstRequest));
                
                itShouldBehaveLikeWithContext(@"normal request", @"NLOperationGETRewriteURL request", MAKE_CONTEXT(secondRequest));
                
                describe(@"delegate", ^{
                    it(@"should be asked whether to start requests for both URLs", ^{
                        assertThat([delegate shouldBeginRequests], hasItems(url, rewritten, nil));
                    });
                    
                    it(@"should be notified a request with the rewritten URL has been started once", ^{
                        assertThatUnsignedInteger([[delegate beganRequests] countForObject:rewritten], is(equalToUnsignedInteger(1)));
                    });
                    
                    it(@"should not be notified a request with the URL to rewrite has finished", ^{
                        assertThat([delegate finishedRequests], isNot(hasItem(url)));
                    });
                    
                    it(@"should be notified a request with the rewritten URL has finished once", ^{
                        assertThatUnsignedInteger([[delegate finishedRequests] countForObject:rewritten], is(equalToUnsignedInteger(1)));
                    });
                    
                    it(@"should be notified a request with both URLs return its resources", ^{
                        assertThat([delegate returningRequests], hasItems(url, rewritten, nil));
                    });
                    
                    it(@"should have approved the URL rewritting", ^{
                        assertThatBool([delegate rewriteApproved], is(equalToBool(YES)));
                    });
                    
                    it(@"should be notified all requests did end", ^{
                        assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                    });
                });
            });
            
            describe(@"when the request URLs are different", ^{
                NSURL *url1 = [NSURL URLWithString:NLOperationGETRawData];
                NSURL *url2 = [NSURL URLWithString:NLOperationGETRewrittenURL];
                
                beforeEach(^{
                    loadRequestWithURL(url1, firstRequest);
                    loadRequestWithURL(url2, secondRequest);
                    CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
                });
                
                itShouldBehaveLikeWithContext(@"first request", @"NLOperationGETRawData request", MAKE_CONTEXT(firstRequest));
                
                itShouldBehaveLikeWithContext(@"second request", @"NLOperationGETRewriteURL request", MAKE_CONTEXT(secondRequest));
                
                describe(@"delegate", ^{
                    it(@"should be asked whether to start requests for both URLs", ^{
                        assertThat([delegate shouldBeginRequests], hasItems(url1, url2, nil));
                    });
                    
                    it(@"should be notified requests have started for both URLs", ^{
                        assertThat([delegate beganRequests], hasItems(url1, url2, nil));
                    });
                    
                    it(@"should be notified requests have finished for both URLs", ^{
                        assertThat([delegate finishedRequests], hasItems(url1, url2, nil));
                    });
                    
                    it(@"should be notified requests return resources for both URLs", ^{
                        assertThat([delegate returningRequests], hasItems(url1, url2, nil));
                    });
                    
                    it(@"should be notified all requests did end", ^{
                        assertThatBool([delegate didEndAllReceived], is(equalToBool(YES)));
                    });
                });
            });
        });
    });
    
    describe(@"Use data parser", ^{
        it(@"should return raw data", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETRawData];
            
            __block id data = nil;
            
            [dataProvider loadResourcesForURL:url options:nil completion:^(id resources, NSDictionary *extraInfo, NSError *error) { data = [resources retain]; }];
            CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
            
            assertThat(data, is(equalTo([NLOperationGETRawData dataUsingEncoding:NSUTF8StringEncoding])));
        });
        
        it(@"should match the MIME type of the response", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETParsedMIMEType];
            
            __block id data = nil;
            
            [dataProvider loadResourcesForURL:url options:nil completion:^(id resources, NSDictionary *extraInfo, NSError *error) { data = [resources retain]; }];
            CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
            
            assertThat(data, is(equalTo(NLOperationGETParsedMIMEType)));
        });
        
        it(@"should match the URL extension", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETParsedURLExt];
            
            __block id data = nil;
            
            [dataProvider loadResourcesForURL:url options:nil completion:^(id resources, NSDictionary *extraInfo, NSError *error) { data = [resources retain]; }];
            CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
            
            assertThat(data, is(equalTo(NLOperationGETParsedURLExt)));
        });
        
        it(@"should match the suggested file name of the response", ^{
            NSURL *url = [NSURL URLWithString:NLOperationGETParsedSuggest];
            
            __block id data = nil;
            
            [dataProvider loadResourcesForURL:url options:nil completion:^(id resources, NSDictionary *extraInfo, NSError *error) { data = [resources retain]; }];
            CFRunLoopRunInMode((CFStringRef)NSDefaultRunLoopMode, 10.0, NO);
            
            assertThat([data autorelease], is(equalTo(NLOperationGETParsedSuggest)));
        });
    });
}

@end


@implementation NLRemoteDataProviderDelegate
@synthesize rewriteApproved, didEndAllReceived, shouldBeginRequests, returningRequests, beganRequests, finishedRequests, cancelledRequests;

- (id)init
{
    self = [super init];
    if(self != nil)
    {
        shouldBeginRequests = [[NSCountedSet alloc] init];
        beganRequests       = [[NSCountedSet alloc] init];
        finishedRequests    = [[NSCountedSet alloc] init];
        returningRequests   = [[NSCountedSet alloc] init];
        cancelledRequests   = [[NSCountedSet alloc] init];
    }
    return self;
}

- (void)dealloc
{
    [shouldBeginRequests release];
    [beganRequests       release];
    [finishedRequests    release];
    [returningRequests   release];
    [cancelledRequests   release];
    [super               dealloc];
}

- (NSURL *)remoteDataProvider:(NLRemoteDataProvider *)sender requestURLForResourceURL:(NSURL *)resourceURL;
{
    if([[resourceURL absoluteString] isEqualToString:NLOperationGETRewriteURL])
    {
        [self setRewriteApproved:YES];
        return [NSURL URLWithString:NLOperationGETRewrittenURL];
    }
    
    return resourceURL;
}

- (void)remoteDataProvider:(NLRemoteDataProvider *)sender didBeginRequestForURL:(NSURL *)URLRequest;  { [beganRequests     addObject:URLRequest]; }
- (void)remoteDataProvider:(NLRemoteDataProvider *)sender didFinishRequestForURL:(NSURL *)URLRequest; { [finishedRequests  addObject:URLRequest]; }
- (void)remoteDataProvider:(NLRemoteDataProvider *)sender didCancelRequestForURL:(NSURL *)URLRequest; { [cancelledRequests addObject:URLRequest]; }
- (void)remoteDataProviderDidFinishAllRequests:(NLRemoteDataProvider *)sender; { [self setDidEndAllReceived:YES]; }

- (BOOL)remoteDataProvider:(NLRemoteDataProvider *)sender shouldBeginRequestForResourceURL:(NSURL *)resourceURL options:(NSDictionary *)options completion:(void(^)(id resources, NSDictionary *extraInfo, NSError *error))completionBlock;
{
    [shouldBeginRequests addObject:resourceURL];
    
    return ![[resourceURL absoluteString] isEqualToString:NLOperationGETDelegatePrevent];
}

- (void)remoteDataProvider:(NLRemoteDataProvider *)sender willReturnResources:(id)resources forResourceURL:(NSURL *)resourceURL withExtraInfo:(NSDictionary *)extraInfo error:(NSError *)error;
{
    [returningRequests addObject:resourceURL];
}

- (BOOL)canParseMIMEType:(NSString *)MIMEType;
{
    return [MIMEType isEqualToString:@"text/ext"];
}

- (BOOL)canParsePathExtension:(NSString *)extension;
{
    return [extension isEqualToString:@"ext"];
}

- (id)parsedObjectForData:(NSData *)data withURL:(NSURL *)aURL options:(NSDictionary *)options error:(NSError **)error;
{
    return [[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease];
}

@end

